# pylint: disable=too-many-lines
from __future__ import annotations

import re
import uuid
from copy import deepcopy
from dataclasses import dataclass
from functools import lru_cache, wraps
from typing import TYPE_CHECKING, Any, Callable, Generator, Iterable, Literal, Type

from slate import Event, Key, Span, Terminal
from slate.span import EMPTY_SPAN
from zenith.markup import FULL_RESET, zml_get_spans, zml_pre_process

from ..enums import Alignment, MouseAction, Overflow, Anchor
from ..frames import Frame, get_frame
from ..state_machine import StateMachine
from ..style_map import StyleMap

if TYPE_CHECKING:
    from ..application import Application, Page
    from .container import Container

__all__ = ["Widget", "widget_types", "handle_mouse_on_children"]

RE_FULL_UNSETTER = re.compile(r"\/(?=\]| )")


def _get_mouse_action_name_options(action: MouseAction) -> tuple[str, ...]:
    if action.value == "hover":
        return (action.value,)

    parts = action.value.split("_")

    full = action.value
    alt_key = "_".join(parts[-2:])
    event = parts[-1]

    if alt_key in ["left", "right"]:
        return (full, alt_key, event)

    return (full, event)


_MOUSE_ACTION_NAME_OPTIONS = {
    action: _get_mouse_action_name_options(action) for action in MouseAction
}


@dataclass
class BoundStyle:
    """A callable object containing both style and fill markup."""

    style: str
    fill: str

    def __call__(self, item: str) -> str:
        return f"[{self.fill}{self.style}]{item}"


def _compute(spec: int | float | None, hint: int) -> int:
    if isinstance(spec, float):
        return int(spec * hint)

    # Auto (fill available)
    if spec is None:
        return hint

    # Static (int)
    return spec


def _overflows(real: int, virt: int) -> bool:
    """Determines whether the given real and virtual dimensions overflow."""

    return real / max(virt, 1) <= 1.0


def to_widget_space(pos: tuple[int, int], widget: Widget) -> tuple[int, int]:
    """Computes a position as an offset from the widget's origin.

    This doesn't account for positions _outside_ of a widget.
    """

    x_offset = pos[0] - widget.position[0] - len(widget.frame.left) - 1
    y_offset = pos[1] - widget.position[1] - len(widget.frame.top) - 1

    x_offset += widget.scroll[0]
    y_offset += widget.scroll[1]

    return (x_offset, y_offset)


def wrap_callback(callback: Callable[[], Any]) -> Callable[[Widget], bool]:
    """Creates a wrapper for empty widget callbacks.

    This lets us maintain the callback's __name__, as opposed to doing
    `lambda _: callback()`
    """

    @wraps(callback)
    def _inner(_: Widget) -> bool:
        value = callback()

        if isinstance(value, bool):
            return value

        return True

    return _inner


class Widget:  # pylint: disable=too-many-instance-attributes,too-many-public-methods
    """This is a docstring."""

    width: int | float | None
    """The hint used by the widget to calculate it's width. See dimension hints."""

    height: int | float | None
    """The hint used by the widget to calculate it's height. See dimension hints."""

    computed_width: int
    """The current width of the widget, calculated from `width`'s hint."""

    computed_height: int
    """The current height of the widget, calculated from `height`'s hint."""

    scroll: tuple[int, int]
    """The widget's (horizontal, vertical) scrolling offset."""

    position: tuple[int, int]
    """The widget's (horizontal, vertical) position."""

    app: "Application"
    """The most recently run Application instance."""

    scroll_step: int = 1
    consumes_mouse: bool = False

    state_machine = StateMachine(
        states=("idle", "hover", "selected", "active", "disabled"),
        transitions={
            "idle": {
                "DISABLED": "disabled",
                "HOVERED": "hover",
                "SELECTED": "selected",
                "CLICKED": "active",
            },
            "hover": {
                "DISABLED": "disabled",
                "RELEASED": "idle",
                "SELECTED": "selected",
                "CLICKED": "active",
            },
            "selected": {
                "DISABLED": "disabled",
                "UNSELECTED": "idle",
                "CLICKED": "active",
            },
            "active": {
                "DISABLED": "disabled",
                "RELEASED": "hover",
                "RELEASED_KEYBOARD": "selected",
            },
            "disabled": {
                "ENABLED": "idle",
            },
            "/": {
                "SUBSTATE_ENTER_SCROLLING_X": "/scrolling_x",
                "SUBSTATE_ENTER_SCROLLING_Y": "/scrolling_y",
            },
            "/scrolling_x": {
                "SUBSTATE_EXIT_SCROLLING_X": "/",
            },
            "/scrolling_y": {
                "SUBSTATE_EXIT_SCROLLING_Y": "/",
            },
        },
    )
    """The state machine keeps track of the current state of the widget.

    It has a tuple of possible states, and a mapping of:

        {current_state: {action: resulting_state, ...}

    ...that are used for state-transitions.
    """

    style_map = StyleMap(
        {
            "idle": {
                "fill": "@.panel1-3",
                "frame": ".panel1+1",
                "content": ".text-1",
                "scrollbar_x": ".panel1+3*0.2",
                "scrollbar_y": ".panel1+3*0.2",
            },
            "hover": {
                "fill": "@.panel1-3",
                "frame": ".panel1+1",
                "content": ".text-1",
                "scrollbar_x": ".panel1+3*0.2",
                "scrollbar_y": ".panel1+3*0.2",
            },
            "selected": {
                "fill": "@.panel1-3",
                "frame": ".panel1+1",
                "content": ".text+1",
                "scrollbar_x": ".panel1+3*0.2",
                "scrollbar_y": ".panel1+3*0.2",
            },
            "active": {
                "fill": "@.panel1-3",
                "frame": ".panel1+1",
                "content": ".text-1",
                "scrollbar_x": ".panel1+3*0.2",
                "scrollbar_y": ".panel1+3*0.2",
            },
            "disabled": {
                "fill": "",
                "frame": ".panel1",
                "content": ".panel1+2",
                "scrollbar_x": ".panel1",
                "scrollbar_y": ".panel1",
            },
            "/scrolling_x": {
                "scrollbar_x": ".primary",
            },
            "/scrolling_y": {
                "scrollbar_y": ".primary",
            },
        }
    )
    """The style map is the lookup table for the widget's styles at certain states."""

    rules = ""
    """YAML rules for this widget, applied only when added as part of an Application."""

    _frame: Frame
    layer: int = 0

    def __init__(
        self,
        eid: str | None = None,
        group: str | None = None,
        groups: tuple[str, ...] = tuple(),
        disabled: bool = False,
    ) -> None:
        """Initializes a Widget.

        Args:
            eid: The id for this widget. Defaults to a UUID.
            group: If set, `groups` is overwritten as `(group,)`.
            groups: The initial groups this widget will belong to.
        """

        self.eid = eid or str(uuid.uuid4())
        if group is not None:
            groups = (group,)
        self.groups = tuple(groups)
        self.position = (0, 0)
        self.state_machine = deepcopy(self.state_machine)
        self.parent: "Container" | "Page" | None = None
        self.disabled = disabled

        self.width: int | float | None = None
        self.height: int | float | None = None
        self.offset: tuple[int, int] = (0, 0)

        # These conversions are handled in their properties
        self.frame = get_frame(None)  # type: ignore
        self.alignment = ("start", "start")  # type: ignore
        self.overflow = ("hide", "hide")  # type: ignore
        self.anchor = Anchor.NONE
        self.palette: str = "main"

        self._virtual_width = 0
        self._virtual_height = 0
        self._last_query: str | None = None
        self._selected_index: int | None = None
        self._selected: Widget | None = None

        self._clip_start: tuple[int, int] = (0, 0)
        self._clip_end: tuple[int, int] = (0, 0)
        self._bindings: dict[str, Event] = {}

        self.computed_width = 1
        self.computed_height = 1
        # Used for width: shrink/fill
        self.width_offset = 0
        # Used for height: shrink/fill
        self.height_offset = 0

        self.pre_content: Event[Widget] = Event("pre content")
        self.on_content: Event[Widget] = Event("post content")

        self.pre_build: Event[Widget] = Event("pre build")
        self.on_build: Event[Widget] = Event("post build")

        self._scrollbar_x = None
        self._scrollbar_y = None

        self._scrolling_x = False
        self._scrolling_y = False

        self._scrollbar_corner_fill = None
        self._scroll = (0, 0)

        self._mouse_target: Widget | None = None
        self._hover_target: Widget | None = None

        self.setup()

        widget_types[type(self).__name__] = type(self)

    def __str__(self) -> str:
        return self.as_query()

    @property
    def terminal(self) -> Terminal | None:
        """Returns the app's terminal."""

        if self.app is None:
            return None

        return self.app.terminal

    @property
    def scrollbar_x(self) -> "Slider":
        from .slider import Slider

        if self._scrollbar_x is None:
            self._scrollbar_x = Slider(groups=("-scroll", "-scroll-x"))

            self._scrollbar_x.parent = self
            self._scrollbar_x.on_change += (
                lambda val: self.scroll_to(x=int(val * self._virtual_width)) or True
            )

        return self._scrollbar_x

    @property
    def scrollbar_y(self) -> "Slider":
        from .slider import VerticalSlider

        if self._scrollbar_y is None:
            self._scrollbar_y = VerticalSlider(groups=("-scroll", "-scroll-y"))

            self._scrollbar_y.parent = self
            self._scrollbar_y.on_change += (
                lambda val: self.scroll_to(y=int(val * self._virtual_height)) or True
            )

        return self._scrollbar_y

    @property
    def scrollbar_corner_fill(self) -> Widget:
        if self._scrollbar_corner_fill is None:
            from .text import Text

            self._scrollbar_corner_fill = Text(" ")
            self._scrollbar_corner_fill.parent = self

        return self._scrollbar_corner_fill

    @property
    def scroll(self) -> tuple[int, int]:
        return self._scroll

    @scroll.setter
    def scroll(self, new: tuple[int, int]) -> None:
        self._scroll = new

        if self._virtual_width > 0:
            self.scrollbar_x.value = new[0] / self._virtual_width
            self.scrollbar_y.value = new[1] / self._virtual_height

    @property
    def state(self) -> str:
        """Returns the current state of the widget."""

        return self.state_machine()

    @property
    def styles(self) -> dict[str, BoundStyle]:
        """Returns a dictionary of current state's style keys to bound styles.

        Note that the `fill` style is inserted into every other style, and is
        stored under the `_fill` key for special circumstances where you may
        need to reference it.
        """

        palette = self.palette

        def _fill_palette(style: str) -> str:
            words = []

            for word in style.split(" "):
                if not (word.startswith(".") or word.startswith("@.")):
                    words.append(word)
                    continue

                alpha = ""

                if "*" in word:
                    word, alpha = word.split("*")
                    alpha = "*" + alpha

                words.append(word.replace(".", palette + ".", 1) + alpha)

            return " ".join(words)

        styles = self.style_map[self.state.split("/")[0]].copy()

        if "/" in self.state:
            key = "/" + self.state.split("/")[-1]

            if key in self.style_map:
                styles |= self.style_map[key].items()

        output = {}

        fill = _fill_palette(styles["fill"])

        if fill != "":
            fill += " "

        for key, style in styles.items():
            if key == "fill":
                key = "_fill"

            style = _fill_palette(style)

            output[key] = BoundStyle(style, fill)

        return output

    @property
    def selected(self) -> Widget | None:
        """Returns the currently selected widget."""

        if self._selected_index is None:
            return None

        return self

    @property
    def selected_index(self) -> int | None:
        return self._selected_index

    @property
    def selectable_count(self) -> int:
        return 1 - self.disabled

    @property
    def frame(self) -> Frame:
        """Returns and sets the current frame."""

        return self._frame

    @frame.setter
    def frame(self, new: str | tuple[str, str, str, str] | Type[Frame]) -> None:
        """Sets the frame setting."""

        current = getattr(self, "_frame", None)

        if isinstance(new, tuple):
            sides = tuple(get_frame(side) for side in new)
            self._frame = Frame.compose(sides)  # type: ignore
            return

        if isinstance(new, str) or new is None:
            new = get_frame(new)

        self._frame = new()

    @property
    def alignment(self) -> tuple[Alignment, Alignment]:
        """Returns and sets the current alignment settings.

        Args:
            new: A tuple of either two strings (start, center, end) or two alignment
                constants.
        """

        return self._alignment

    @alignment.setter
    def alignment(self, new: tuple[Alignment, Alignment] | tuple[str, str]) -> None:
        """Sets the alignment setting."""

        if not all(isinstance(item, Alignment) for item in new):
            new = Alignment(new[0]), Alignment(new[1])

        assert isinstance(new[0], Alignment) and isinstance(new[1], Alignment)
        self._alignment = new

    @property
    def overflow(self) -> tuple[Overflow, Overflow]:
        """Returns and sets the current overflow settings.

        Args:
            new: A tuple of either two strings (start, center, end) or two alignment
                constants.
        """

        return self._overflow

    @overflow.setter
    def overflow(self, new: tuple[Overflow, Overflow] | tuple[str, str]) -> None:
        """Sets the overflow setting."""

        if not all(isinstance(item, Overflow) for item in new):
            new = Overflow(new[0]), Overflow(new[1])

        assert isinstance(new[0], Overflow) and isinstance(new[1], Overflow)
        self._overflow = new

    @property
    def anchor(self) -> Anchor:
        """Returns and sets the widget's anchor."""

        return self._anchor

    @anchor.setter
    def anchor(self, new: str | Anchor) -> None:
        """Sets the widget's anchor."""

        if isinstance(new, str):
            new = Anchor(new)

        self._anchor = new

    @property
    def _framed_width(self) -> int:
        """Gets the widget's width excluding its frame."""

        return max(self.computed_width - self.frame.width, 0)

    @property
    def _framed_height(self) -> int:
        """Gets the widget's height excluding its frame."""

        return max(self.computed_height - self.frame.height, 0)

    @property
    def disabled(self) -> bool:
        """Returns the widget's disabled state."""

        return self._disabled

    @disabled.setter
    def disabled(self, value: bool) -> None:
        """Sets the widget's disabled state."""

        self._disabled = value

        if value:
            self.state_machine.apply_action("DISABLED")
        else:
            self.state_machine.apply_action("ENABLED")

    @property
    def clipped_position(self) -> tuple[int, int]:
        """Returns the current position offset by the current clip start position."""

        return (
            self.position[0] + self._clip_start[0],
            self.position[1] + self._clip_start[1],
        )

    @property
    def inner_rect(self) -> tuple[tuple[int, int], tuple[int, int]]:
        """Returns the inner rect of this widget."""

        frame = self.frame
        frame_bottom_raw = self.frame.bottom != ""
        frame_right_raw = self.frame.right != ""

        frame_left = (frame.left != "") * (self._clip_start[0] == 0)
        frame_right = frame_right_raw * (self._clip_end[0] == 0)
        frame_top = (frame.top != "") * (self._clip_start[1] == 0)
        frame_bottom = frame_bottom_raw * (self._clip_end[1] == 0)

        # Only add space for bars if they are clipped
        right_bar = self.has_scrollbar(1)
        if self._clip_end[1] > frame_right_raw:
            right_bar = 0

        bottom_bar = self.has_scrollbar(0)
        if self._clip_end[1] > frame_bottom_raw:
            bottom_bar = 0

        return (
            (
                self.position[0] + self._clip_start[0] + frame_left,
                self.position[1] + self._clip_start[1] + frame_top,
            ),
            (
                (
                    self.position[0]
                    + self.computed_width
                    - frame_right
                    - self._clip_end[0]
                    - right_bar
                ),
                (
                    self.position[1]
                    + self.computed_height
                    - frame_bottom
                    - self._clip_end[1]
                    - bottom_bar
                ),
            ),
        )

    @property
    def outer_rect(self) -> tuple[tuple[int, int], tuple[int, int]]:
        """Returns the outer rect of this widget."""

        return self.position, (
            self.position[0] + self.computed_width,
            self.position[1] + self.computed_height,
        )

    @lru_cache(1024)
    def _parse_markup(self, markup: str) -> tuple[Span, ...]:
        """Parses some markup into a span of tuples.

        This also handles (ignores) zenith's FULL_RESET spans, and replaces `/`
        unsetters with `/ {fill_style} {content_style}`.
        """

        fill = self.styles["_fill"]

        if isinstance(self.parent, Widget) and fill.style == "":
            fill = self.parent.styles["_fill"]

        content_style = fill.style + " " + self.styles["content"].style

        # Replace full unsetters with full unsetter + content style
        markup = RE_FULL_UNSETTER.sub("/ " + content_style, markup)
        markup = zml_pre_process(markup)

        return tuple(span for span in zml_get_spans(markup) if span is not FULL_RESET)

    def _update_scrollbars(self, width: int, height: int) -> None:
        def _get_size(computed: int, virtual: int, framed: int) -> int:
            return int(computed * (framed / (virtual or framed)))

        self.scrollbar_x.compute_dimensions(width, 1)
        self.scrollbar_y.compute_dimensions(1, height)

        frame_left = self.frame.left != ""
        frame_top = self.frame.top != ""
        frame_right = self.frame.right != ""
        frame_bottom = self.frame.bottom != ""

        clip_start = list(self._clip_start)
        clip_end = list(self._clip_end)

        [start_x, start_y], [end_x, end_y] = self.outer_rect

        clip_start[0] = max(0, clip_start[0] - frame_left)
        clip_start[1] = max(0, clip_start[1] - frame_top)

        start_x += self.frame.left != ""
        start_y += self.frame.top != ""

        clip_end[0] = max(0, clip_end[0] - frame_right)
        clip_end[1] = max(0, clip_end[1] - frame_bottom)

        end_x -= frame_right + 1
        end_y -= frame_bottom + 1

        self.scrollbar_x.position = (start_x, end_y)
        self.scrollbar_x.clip(
            (clip_start[0], max(0, clip_start[1] - height)),
            clip_end
        )

        self.scrollbar_y.position = (end_x, start_y)
        self.scrollbar_y.clip(
            (max(0, clip_start[0] - width), clip_start[1]),
            clip_end
        )

        self.scrollbar_corner_fill.position = (end_x, end_y)
        self.scrollbar_corner_fill.clip(clip_start, clip_end)

        self.scrollbar_x.cursor_size = _get_size(
            self.computed_width, self._virtual_width, width
        )
        self.scrollbar_y.cursor_size = _get_size(
            self.computed_height, self._virtual_height, height
        )

    def _apply_frame(  # pylint: disable=too-many-locals
        self, lines: list[tuple[Span, ...]], width: int
    ) -> None:
        """Adds frame characters around the given lines."""

        def _style(item, outer: bool = False) -> tuple[Span, ...]:
            style = deepcopy(self.styles["frame"])

            if outer:
                if self.parent is not None and hasattr(self.parent, "styles"):
                    style.style = self.parent.styles["frame"].fill + " " + style.style

            return self._parse_markup(style(item))

        left_top, right_top, right_bottom, left_bottom = self.frame.corners
        left, top, right, bottom = self.frame.borders

        h_outer = self.frame.outer_horizontal
        v_outer = self.frame.outer_vertical
        c_outer = self.frame.outer_corner

        for i, line in enumerate(lines):
            lines[i] = (
                *_style(left, outer=v_outer),
                *line,
                *_style(right, outer=v_outer),
            )

        if left_top + top + right_top != "":
            lines.insert(
                0,
                (
                    *_style(left_top or (left != "") * top, outer=c_outer),
                    *_style(top * width, outer=h_outer),
                    *_style(right_top or (right != "") * top, outer=c_outer),
                ),
            )

        if left_bottom + bottom + right_bottom != "":
            lines.append(
                (
                    *_style(left_bottom or (left != "") * bottom, outer=c_outer),
                    *_style(bottom * width, outer=h_outer),
                    *_style(right_bottom or (right != "") * bottom, outer=c_outer),
                )
            )

    def _horizontal_align(self, line: tuple[Span, ...], width: int) -> tuple[Span, ...]:
        """Aligns a tuple of spans horizontally, using `self.alignment[0]`."""

        alignment = self.alignment[0]

        width = max(width, self._virtual_width)
        length = sum(len(span.text) for span in line)
        diff = width - length

        style = self.styles["content"]

        if line in [tuple(), (Span(""),)]:
            return self._parse_markup(style(diff * " "))

        if alignment is Alignment.START:
            return (*line[:-1], line[-1].mutate(text=line[-1].text + diff * " "))

        if alignment is Alignment.CENTER:
            end, extra = divmod(diff, 2)
            start = end + extra

            # Return padding as a single span instead of (start, content, end)
            # when content is " " (filler lines)
            if "".join(span.text for span in line) == " ":
                return self._parse_markup(style((start + end) * " "))

            return (
                *self._parse_markup(style(start * " ")),
                line[0],
                *line[1:],
                *self._parse_markup(style(end * " ")),
            )

        if alignment is Alignment.END:
            span = line[0]

            return (span.mutate(text=diff * " " + span.text), *line[1:])

        raise NotImplementedError(f"Unknown alignment {alignment!r}.")

    def _vertical_align(self, lines: list[tuple[Span, ...]], height: int) -> None:
        """Aligns a list of tuples of spans vertically, using `self.alignment[1]`.

        Note that this mutates `lines`.
        """

        alignment = self.alignment[1]

        available = height - len(lines)
        filler = self._parse_markup(self.styles["content"](" "))

        if alignment is Alignment.START:
            lines.extend([filler] * available)
            return

        if alignment is Alignment.CENTER:
            top, extra = divmod(available, 2)
            bottom = top + extra

            for _ in range(top):
                lines.insert(0, filler)

            lines.extend([filler] * bottom)
            return

        for _ in range(available):
            lines.insert(0, filler)

    @lru_cache(1024)
    def _slice_line(
        self, line: tuple[Span, ...], start: int, end: int
    ) -> tuple[Span, ...]:
        """Slices a line into the given width."""

        if end is None:
            end = len(line) - 1

        width = end - start

        line_list = []
        before_start = 0
        occupied = 0

        # Slice to start
        for span in line:
            length = len(span)

            before_start += length
            if before_start < start:
                continue

            new = span[max(start - (before_start - length), 0) :]
            length = len(new)

            if length == 0:
                continue

            line_list.append(new)

            occupied += length

            if occupied > width:
                break

        width_diff = max(occupied - width, 0)
        empty = []

        # Slice from end
        if width_diff > 0 and len(line_list) > 0:
            for i, span in enumerate(reversed(line_list)):
                new = span[:-width_diff]
                width_diff -= len(span) - len(new)

                if len(new) == 0:
                    empty.append(-i - 1)
                else:
                    line_list[-i - 1] = new

                if width_diff <= 0:
                    break

        # Remove 0 length spans
        for offset, i in enumerate(empty):
            line_list.pop(i - offset)

        if not line_list:
            line_list.extend(self._parse_markup(self.styles["content"](" ")))

        occupied = sum(len(span) for span in line_list)

        if occupied < width:
            suffix = line_list[-1]

            line_list[-1] = suffix.mutate(text=suffix.text + (width - occupied) * " ")

        return tuple(line_list)

    def _apply_mouse_state(self, action: MouseAction) -> None:
        """Applies a state change action based on mouse input."""

        value = action.value

        if "click" in value:
            self.state_machine.apply_action("CLICKED")
            return

        if "release" in value:
            self.state_machine.apply_action("RELEASED")

            if self._selected_index is not None:
                self.state_machine.apply_action("SELECTED")

            return

        if "hover" in value:
            self.state_machine.apply_action("HOVERED")
            return

    def _compute_shrink_width(self) -> int:
        """Computes the minimum width this widget's content takes up.

        Used for `width = -1`.
        """

        raise NotImplementedError(f"widget {self!r} does not implement shrink width.")

    def _compute_shrink_height(self) -> int:
        """Computes the minimum height this widget's content takes up.

        Used for `height = -1`.
        """

        raise NotImplementedError(f"widget {self!r} does not implement shrink height.")

    def setup(self) -> None:
        """Use this to do simple setup actions without overriding __init__."""

    def serialize(self) -> dict[str, Any]:
        """Returns a dictionary to represent the dynamic value of the widget."""

        return {}

    def update(
        self,
        attrs: dict[str, Any],
        style_map: dict[str, str],
    ) -> None:
        """Updates a widget with new setttings.

        Args:
            attrs: A dictionary of the widget's fields to values. Fields starting with
                `_` are disallowed, as are fields the widget doesn't already have.
            style_map: A dictionary of style keys to markup that gets merged onto our
                `style_map`'s current state, i.e.:

                    `self.style_map | {self.state: style_map}`
        """

        def _parse_offset(number: str) -> int | float:
            if number == "":
                return 0

            if number.lstrip("-+").replace(".", "", 1).isdigit():
                num = float(number)

                if num.is_integer():
                    return int(num)

                return num

            raise ValueError(
                f"can't parse {key} offset {number!r} "
                + str(value.lstrip("-+").replace(".", "", 1))
            )

        keys = dir(self)

        for key, value in attrs.items():
            if key.startswith("_"):
                raise ValueError(f"cannot set non-public attr {key!r}")

            if key not in keys:
                raise ValueError(f"cannot set non-existant attr {key!r}")

            if key in ["width", "height"] and isinstance(value, str):
                if value.startswith("fill"):
                    offset = _parse_offset(value[len("fill") :])
                    value = 1.0

                    setattr(self, f"{key}_offset", offset)

                elif value.startswith("shrink"):
                    offset = _parse_offset(value[len("shrink") :])
                    value = -1

                    setattr(self, f"{key}_offset", offset)

            setattr(self, key, value)

        self.style_map = self.style_map | {self.state: style_map}

    def as_query(self, state: bool = False) -> str:
        """Returns the widget as the most specific selectable query.

        Args:
            state: Include state suffix onto the query.
        """

        query = type(self).__name__

        if self.eid is not None:
            query += f"#{self.eid}"

        for variant in self.groups:
            query += f".{variant}"

        if state:
            query += f"/{self.state}"

        return query

    def query_changed(self) -> bool:
        """Returns whether the result of `as_query` has changed since last call."""

        query = self.as_query(state=True)
        value = query != self._last_query

        self._last_query = query
        return value

    def is_fill_width(self) -> bool:
        """Shorthand for `width is None`."""

        return self.width is None

    def is_fill_height(self) -> bool:
        """Shorthand for `height is None`."""

        return self.height is None

    def scroll_to(self, x: int | None = None, y: int | None = None) -> None:
        """Scrolls the widget.

        Both arguments may be given in one of four forms:

        - `0`: scroll to the start of the widget on the given axis
        - `-1`: scroll to the end of the widget on the given axis
        - `0<` : scroll to the offset on the given axis
        - `None`: do nothing for the given axis.

        Args:
            x: The x scroll value.
            y: The y scroll value.
        """

        new_x = x if x is not None else self.scroll[0]

        if new_x == -1:
            new_x = self._virtual_width

        new_y = y if y is not None else self.scroll[1]

        if new_y == -1:
            new_y = self._virtual_height

        self.scroll = (new_x, new_y)

    def has_scrollbar(self, index: Literal[0, 1]) -> bool:
        """Returns whether the given dimension should display a scrollbar.

        Args:
            index: The axis to check. 0 for horizontal scrolling, 1 for vertical.
        """

        overflow = self.overflow[index]

        if overflow is Overflow.SCROLL:
            return True

        if overflow is Overflow.HIDE:
            return False

        real, virt = [
            (self._framed_width, self._virtual_width),
            (self._framed_height, self._virtual_height),
        ][index]

        return virt > real

    def drawables(self) -> Iterable[Widget]:
        """Yields all contained widgets that should be drawn."""

        yield self

        x_bar = self.has_scrollbar(0)
        y_bar = self.has_scrollbar(1)

        if x_bar:
            yield self.scrollbar_x

        if y_bar:
            yield self.scrollbar_y

        if x_bar and y_bar:
            yield self.scrollbar_corner_fill

    def clip(self, start: tuple[int, int], end: tuple[int, int]) -> None:
        """Sets the clipping rectangle's coordinates."""

        self._clip_start = start
        self._clip_end = end

    def add_group(self, target: str) -> None:
        """Adds a group to the widget's groups."""

        self.groups = tuple(list(self.groups) + [target])

    def remove_group(self, target: str) -> None:
        """Removes a group from the widget's groups."""

        self.groups = tuple(group for group in self.groups if group != target)

    def toggle_group(self, target: str) -> bool:
        """Toggles a group in the widget's groups."""

        if target in self.groups:
            self.remove_group(target)
            return False

        self.add_group(target)
        return True

    def select(self, index: int | None = None) -> int | None:
        if index == 0 or self.selectable_count == 0:
            return index

        if index is None:
            self._selected_index = None
            self.state_machine.apply_action("UNSELECTED")
            return index

        index -= 1

        if index == 0:
            self._selected_index = index
            self.state_machine.apply_action("SELECTED")

        else:
            self._selected_index = None
            self.state_machine.apply_action("UNSELECTED")

        return index

    def hide(self) -> None:
        """Adds the `hidden` group onto the widget."""

        self.add_group("hidden")

    def show(self) -> None:
        """Remove the `hidden` group from the widget."""

        self.remove_group("hidden")

    def contains(self, position: tuple[int, int]) -> bool:
        """Determines whether this widget contains the given position."""

        rect = self.position, (
            self.position[0] + self.computed_width,
            self.position[1] + self.computed_height,
        )

        (left, top), (right, bottom) = rect

        return left <= position[0] <= right and top <= position[1] <= bottom

    def move_to(self, x: int, y: int) -> None:
        """Moves the widget to the given position."""

        self.position = x, y

    def move_by(self, x: int, y: int) -> None:
        """Moves the widget by the given changes."""

        self.move_to(self.position[0] + x, self.position[1] + y)

    def remove_from_parent(self) -> None:
        """Removes this widget from its parent."""

        if self.parent is None:
            return

        self.parent.remove(self)

    def bind(self, key: str, callback: Callable[[Widget], Any]) -> Event:
        """Binds a key to the given callback.

        If a binding already exists, the given callback is added to the Event.
        """

        if key not in self._bindings:
            self._bindings[key] = Event(f"on key {key}")

        event = self._bindings[key]
        event += callback

        return event

    def unbind(self, key: str) -> Event:
        """Removes all bindings for a given key."""

        if key not in self._bindings:
            raise ValueError(f"can't remove non-existant binding for key {key!r}.")

        event = self._bindings[key]
        del self._bindings[key]

        return event

    def handle_keyboard(self, key: Key) -> bool:
        """Handles a keyboard event.

        Returns:
            Whether the event should continue to bubble upwards. Return `True` if the
            event was already handled.
        """

        if self.disabled:
            return False

        result = False

        for value in key:
            if value in self._bindings:
                if self._bindings[value](self):
                    result |= True

            if result:
                break

        return result

    def handle_mouse(self, action: MouseAction, position: tuple[int, int]) -> bool:
        """Handles a mouse event.

        Returns:
            Whether the event should continue to bubble upwards. Return `True` if the
            event was already handled.
        """

        if self.disabled:
            return False

        self._apply_mouse_state(action)

        bars = []

        if self.has_scrollbar(0):
            bars.append(self.scrollbar_x)

        if self.has_scrollbar(1):
            bars.append(self.scrollbar_y)

        result, mouse_target, hover_target = handle_mouse_on_children(
            action,
            position,
            self._mouse_target,
            self._hover_target,
            # TODO: We might be able to get away with just providing both regardless.
            bars,
        )

        self._mouse_target = mouse_target
        self._hover_target = hover_target

        if result:
            return True

        if "scroll" in action.value:
            can_scroll_x, can_scroll_y = self.has_scrollbar(0), self.has_scrollbar(1)

            if can_scroll_x:
                forward = [MouseAction.SCROLL_LEFT, MouseAction.SHIFT_SCROLL_UP]
                backward = [MouseAction.SCROLL_RIGHT, MouseAction.SHIFT_SCROLL_DOWN]

                if action in forward and self.scroll[0] > 0:
                    self.scroll = (self.scroll[0] - self.scroll_step, self.scroll[1])
                    return True

                if action in backward and self.scroll[0] + self.computed_width - can_scroll_y <= self._virtual_width:
                    self.scroll = (self.scroll[0] + self.scroll_step, self.scroll[1])
                    return True

            if can_scroll_y:
                if action is MouseAction.SCROLL_UP and self.scroll[1] > 0:
                    self.scroll = (self.scroll[0], self.scroll[1] - self.scroll_step)
                    return True

                if action is MouseAction.SCROLL_DOWN and self.scroll[1] + self.computed_height - can_scroll_x < self._virtual_height:
                    self.scroll = (self.scroll[0], self.scroll[1] + self.scroll_step)
                    return True

        for name in _MOUSE_ACTION_NAME_OPTIONS[action]:
            if (handle := getattr(self, f"on_{name}", None)) is not None:
                return handle(action, position)

        # TODO: Scroll propagation algorithm:
        #
        #       on start scroll:
        #           is scrollable? return True
        #           return False
        #
        #       while continuously scrolling:
        #           return True

        return False

    def compute_dimensions(self, available_width: int, available_height: int) -> None:
        """Computes width & height based on our specifications and the given space."""

        width_offset = _compute(self.width_offset, available_width)

        if self.width == -1:
            self.computed_width = self._compute_shrink_width() + self.frame.width
        else:
            self.computed_width = _compute(self.width, available_width)

        height_offset = _compute(self.height_offset, available_height)

        if self.height == -1:
            self.computed_height = self._compute_shrink_height() + self.frame.height
        else:
            self.computed_height = _compute(self.height, available_height)

        self.computed_width += self.width_offset
        self.computed_height += self.height_offset

    def get_content(self) -> list[str]:
        """Gets the dynamic content for this widget."""

        raise NotImplementedError

    def build(
        self, *, virt_width: int | None = None, virt_height: int | None = None
    ) -> list[tuple[Span, ...]]:
        """Builds the strings that represent the widget."""

        self.pre_build(self)

        width = self._framed_width
        height = self._framed_height

        def _clamp_scrolls() -> tuple[int, int]:
            x_bar = width < self._virtual_width
            y_bar = height < self._virtual_height

            return (
                max(min(self.scroll[0], self._virtual_width - width + x_bar), 0),
                max(min(self.scroll[1], self._virtual_height - height + y_bar), 0),
            )

        self.scroll = _clamp_scrolls()

        self.pre_content(self)

        content_style = self.styles["content"]
        lines: list[tuple[Span, ...]] = [
            self._parse_markup(content_style(line)) for line in self.get_content()
        ]

        self.on_content(self)

        self._virtual_height = virt_height or len(lines) or 1
        self._virtual_width = virt_width or max(
            (sum(len(span) for span in line) for line in lines), default=1
        )

        # Clip into vertical size
        if self._virtual_height > height:
            lines = lines[self.scroll[1] : self.scroll[1] + height]

            if len(lines) < height:
                lines.extend([(EMPTY_SPAN,)] * (height - len(lines)))

        # Handle alignment
        self._vertical_align(lines, height)

        for i, line in enumerate(lines):
            lines[i] = self._horizontal_align(line, width)

        # Clip into horizontal size
        lines = [
            self._slice_line(line,
                self.scroll[0],
                self.scroll[0] + width
            )
            for line in lines
        ]

        self._apply_frame(lines, width)

        if len(lines) > self.computed_height:
            lines = lines[: self.computed_height]

        # TODO: This double-slice could be done in one:
        #
        #         slice scroll + clip (offset by frame?)
        #         remove frame if clipped?

        lines = [
            self._slice_line(line,
                self._clip_start[0],
                self.computed_width - self._clip_end[0]
            )
            for line in lines
        ]

        lines = lines[self._clip_start[1] :]

        if self._clip_end[1] > 0:
            lines = lines[: -self._clip_end[1]]

        both = self.has_scrollbar(0) and self.has_scrollbar(1)
        self._update_scrollbars(width - both, height - both)

        self.on_build(self)

        if (
            self._clip_start[0] + self._clip_end[0] >= self.computed_width
            or self._clip_start[1] + self._clip_end[1] >= self.computed_height
        ):
            return []

        return lines


widget_annotations = Widget.__annotations__
widget_types: dict[str, Type[Widget]] = {"Widget": Widget}


def handle_mouse_on_children(
    action: MouseAction,
    position: tuple[int, int],
    mouse_target: Widget | None,
    hover_target: Widget | None,
    children: Iterable[Widget],
) -> tuple[bool, Widget | None, Widget | None]:
    """Handles the given mouse event on an iterable of widgets.

    This can be used by 'container' type widgets (like `Container`), or meta-level
    management objects (like `Application`) for consistent mouse handling behaviour.

    Args:
        action: The action associated with the event.
        position: The position associated with the event.
        mouse_target: The caller's current mouse target.
        hover_target: The caller's current hover target.
        children: The widgets to execute upon.

    Returns:
        A tuple of:

            (success, mouse_target, hover_target)

        The input state changes based on the following rules:

        - Success is False _only_ if no child contains the given position.
        - Mouse target is...
            - Set to None on release events, or
            - Untouched if the event was not a form of "click" and the current mouse
              target could handle it, or
            - Untouched if the event was captured by another child but it could not
              handle it, or
            - Set to the first child that contained & handled the event.
        - Hover target is...
            - Set to None on release events, or
            - Set to None if a hover event is done outside the current hover target, or
            - Set to the first child that contained & handled a hover event,
            - Untouched otherwise.
    """

    if action is MouseAction.LEFT_RELEASE:
        if mouse_target is not None:
            mouse_target.handle_mouse(action, position)
            mouse_target = None

        if hover_target is not None:
            hover_target.handle_mouse(action, position)
            hover_target = None

        return True, mouse_target, hover_target

    is_hover = action is MouseAction.HOVER
    release = MouseAction.LEFT_RELEASE, position

    if (
        mouse_target is not None
        and "click" not in action.value  # Clicks cannot be done outside of the widget
        and mouse_target.handle_mouse(action, position)
    ):
        return True, mouse_target, hover_target

    scroll_actions = [MouseAction.SCROLL_UP, MouseAction.SCROLL_DOWN, MouseAction.SCROLL_LEFT, MouseAction.SCROLL_RIGHT]

    if hover_target is not None:
        if action in scroll_actions and hover_target.handle_mouse(action, position):
            return True, mouse_target, hover_target

        if is_hover and not hover_target.contains(position):
            hover_target.handle_mouse(*release)
            hover_target = None

    for child in children:
        if not child.contains(position):
            continue

        if is_hover:
            if hover_target is not None and hover_target is not child:
                hover_target.handle_mouse(*release)

            hover_target = child

        if (handled := child.handle_mouse(action, position)) or child.consumes_mouse:
            if mouse_target is not None and mouse_target is not child:
                mouse_target.handle_mouse(*release)

            if child.consumes_mouse and not handled:
                return True, mouse_target, hover_target

            return True, child, hover_target

        break

    return False, mouse_target, hover_target



from slate import Color
from zenith import Palette, zml_alias

palette = Palette("celadon", namespace="main.")
palette.alias()

zml_alias(**{"ui.text": Color((245, 245, 245)).darken(2).hex})  # type: ignore



from __future__ import annotations

import re
from typing import Any

from zenith import zml_wrap

from ..enums import MouseAction
from .widget import Widget

RE_HYPERLINK_MARKUP = re.compile(r"~([^ \]\[]+)")


class Text(Widget):
    """A widget that displays some static text."""

    wrap: bool = False

    rules = """
    Text, Link:
        width: shrink
        height: shrink
    """

    def __init__(self, content: str, **widget_args: Any) -> None:
        super().__init__(**widget_args)

        self.content = content

        self._wrapped_content = []

        def _wrap_content(_: Widget) -> bool:
            if self.wrap:
                self._wrapped_content = zml_wrap(self.content, width=self._framed_width)
            else:
                self._wrapped_content = self.content.splitlines()

            return True

        self.pre_content += _wrap_content

    def _compute_shrink_width(self) -> int:
        return max(
            (
                sum(len(span) for span in self._parse_markup(line))
                for line in self.content.splitlines()
            ),
            default=0,
        )

    def _compute_shrink_height(self) -> int:
        return len(self._wrapped_content)

    @property
    def selectable_count(self) -> int:
        return 0

    def on_click(self, _: MouseAction, __: tuple[int, int]) -> bool:
        """Clicks the first hyperlink within the text's content."""

        if self.app is None:
            return False

        if (mtch := RE_HYPERLINK_MARKUP.search(self.content)) is not None:
            value = mtch[1]

            self.app.route(value)

        # Never return True, as we don't want the parent to select a widget
        return False

    def get_content(self) -> list[str]:
        return self._wrapped_content

    # Proper linebreaking prototype
    #
    # lines = []
    # usable_width = self.width - self.frame.width - 1
    #
    # for line in self.content.splitlines():
    #     length = 0
    #     buff = ""
    #
    #     for char in line:
    #         if char == "[":
    #             in_tag = True
    #
    #         elif char == "]":
    #             in_tag = False
    #
    #         if length < usable_width and char not in ["\n", "\r"]:
    #             if char != "]" and not in_tag:
    #                 length += 1
    #
    #             buff += char
    #             continue
    #
    #         lines.append(buff)
    #         length = 0
    #         buff = ""
    #
    #     lines.append(buff)
    #
    # return lines



from __future__ import annotations

from typing import Any

from slate import Event

from ..enums import MouseAction
from .widget import Widget, to_widget_space, wrap_callback


class Slider(Widget):
    """A clickable and draggable slider."""

    style_map = Widget.style_map | {
        state: {"cursor": ""} for state in Widget.state_machine.states
    }

    on_change: Event
    """Called when the slider's value changes.

    Args:
        Value: The new value of the slider.
    """

    cursor: str = "┃"
    rail: str = "─"

    rules = """
    Slider,VerticalSlider:
        content_style: '.panel1'
        cursor_style: '.panel1+1'

        /selected|hover:
            fill_style: '@.panel1-2'

        /selected|active:
            cursor_style: '.primary'

        /disabled:
            cursor_style: .panel1-1
            content_style: .panel1-1

        .-scroll-x:
            rail: " "
            cursor: "▅"

        .-scroll-y:
            rail: " "
            cursor: "█"

    Slider:
        height: 1
        width: null

    VerticalSlider:
        height: null
        width: 1
    """

    def __init__(
        self,
        *,
        value: float = 0.0,
        scale: float = 1.0,
        precision: int = 1,
        cursor_size: int = 1,
        name: str | None = None,
        **widget_args: Any,
    ) -> None:
        """Initializes a Slider.

        Args:
            value: The internal value (0.0-1.0) to start at.
            scale: The factor to scale the internal value by (`value = _value * scale`).
            precision: The rounding precision used in `increase` & `decrease` to avoid
                floating point errors.
            name: The name used when serializing this widget.
        """

        super().__init__(**widget_args)

        self._value = value
        self._scale = scale

        self.precision = precision
        self.name = name
        self.cursor_size = cursor_size

        self.on_change: Event[float] = Event("slider change")

        # TODO: These can be directional with class-body definitions
        self.bind("right", wrap_callback(self.increase))
        self.bind("left", wrap_callback(self.decrease))

        self._grab_offset = 0

    @property
    def value(self) -> float:
        """Returns the internal state of progress bar scaled to the end factor."""

        return self._value * self._scale

    @value.setter
    def value(self, new: float) -> None:
        self._value = new

    def serialize(self) -> dict[str, Any]:
        if self.name is None:
            raise ValueError(f"slider {self!r} cannot be serialized without a name.")

        return {self.name: self.value}

    # TODO: This docstring fucking sucks lol
    def _get_value(self, offset: int) -> float:
        """Gets the fractional value at the given widget offset."""

        x = to_widget_space((offset, 0), self)[0]
        return max(0.0, min((x - self._grab_offset) / self.computed_width, 1.0))

    def increase(self, amount: float = 0.1) -> bool:
        """Increases the progress bar by the given amount."""

        self._value = min(round(self._value + amount, self.precision), 1.0)
        self.on_change(self.value)

        return True

    def decrease(self, amount: float = 0.1) -> bool:
        """Decreases the progress bar by the given amount."""

        self._value = max(round(self._value - amount, self.precision), 0.0)
        self.on_change(self.value)

        return True

    def on_click(self, _: MouseAction, pos: tuple[int, int]) -> bool:
        cursor = round(self.computed_width * self._value)
        self._grab_offset = to_widget_space(pos, self)[0] - cursor

        # Click outside of cursor bar, so we recenter around it
        if not (0 <= self._grab_offset < cursor + self.cursor_size):
            self._grab_offset = self.cursor_size // 2

        self._value = self._get_value(pos[0])
        self.on_change(self.value)

        return True

    def on_drag(self, _: MouseAction, pos: tuple[int, int]) -> bool:
        self._value = self._get_value(pos[0])
        self.on_change(self.value)

        return True

    def _build(self, dimension: int) -> list[str]:
        start = min(round(dimension * self._value), dimension - self.cursor_size)

        style = self.styles["content"]

        return [
            *[style(self.rail) for _ in range(start)],
            *[self.styles["cursor"](self.cursor) for _ in range(self.cursor_size)],
            *[style(self.rail) for _ in range(dimension - start - self.cursor_size)],
        ]

    def get_content(self) -> list[str]:
        return ["".join(self._build(self.computed_width))]


class VerticalSlider(Slider):
    cursor: str = "━"
    rail: str = "│"

    # TODO: This docstring fucking sucks lol
    def _get_value(self, offset: int) -> float:
        """Gets the fractional value at the given widget offset."""

        x = to_widget_space((0, offset), self)[1]
        return max(0.0, min((x - self._grab_offset) / self.computed_height, 1.0))

    def on_click(self, _: MouseAction, pos: tuple[int, int]) -> bool:
        cursor = round(self.computed_height * self._value)
        self._grab_offset = to_widget_space(pos, self)[1] - cursor

        # Click outside of cursor bar, so we recenter around it
        if not (0 <= self._grab_offset < cursor + self.cursor_size):
            self._grab_offset = self.cursor_size // 2

        self._value = self._get_value(pos[1])
        self.on_change(self.value)

        return True

    def on_drag(self, _: MouseAction, pos: tuple[int, int]) -> bool:
        self._value = self._get_value(pos[1])
        self.on_change(self.value)

        return True

    def get_content(self) -> list[str]:
        return self._build(self.computed_height)



from typing import Iterable

from .widget import Widget
from .container import Tower


class Dialogue(Tower):
    """A full-screen overlay that displays some content."""

    rules = r"""
    Dialogue:
        consumes_mouse: true
        anchor: screen

        width: 1.0
        height: 1.0

        alignment: [center, center]
        fill_style: '@black*0.6'

        \> Tower:
            consumes_mouse: true

            width: 0.6
            height: 0.5

            alignment: [center, start]
            frame: [padded, verticalouter, padded, verticalouter]

        \*> Text.title:
            alignment: [center, center]
            height: 2
            content_style: bold dim

        \*> Text.body:
            height: null
            overflow: [hide, auto]

        \*> Tower.body:
            gap: 1

        \*> Row.input:
            height: 3
            alignment: [center, end]
    """

    def __init__(self, *children: Widget, **widget_args) -> None:
        self.content = Tower(*children)
        super().__init__(**widget_args)

        super().insert(0, self.content)

    def on_click(self, _, __) -> bool:
        """Destroy dialogue if the outside (dim part) is clicked."""

        self.remove_from_parent()
        return True

    # Redirect all operations to .content

    def insert(self, index: int, widget: Widget) -> None:
        self.content.insert(index, widget)

    def append(self, widget: Widget) -> None:
        self.content.append(widget)

    def extend(self, widgets: Iterable[Widget]) -> None:
        self.content.extend(widgets)

    def remove(self, widget: Widget) -> None:
        self.content.remove(widget)

    def pop(self, index: int) -> Widget:
        return self.content.pop(index)

    def clear(self) -> None:
        self.content.clear()

    def update_children(self, widgets: Iterable[Widget]) -> None:
        self.content.update_children(widgets)

    def replace(self, current: Widget, new: Widget, *, offset: int = 0) -> None:
        self.content.replace(current, new, offset=offset)



from __future__ import annotations

from typing import Any, Callable

from slate import Event

from ..enums import MouseAction
from .widget import Widget, wrap_callback


class Button(Widget):
    """A pressable/clickable button."""

    on_submit: Event
    """Called when the Button is clicked/pressed using a primary input (mouse1, return).

    Args:
        self: The instance that sent the event.
    """

    rules = """
    Button:
        content_style: dim
        alignment: [center, center]

        width: shrink
        height: 1

        frame: [heavy, null, heavy, null]

        /idle:
            fill_style: '@.panel1'
            frame_style: .primary+1

        /hover:
            fill_style: '@.panel1+1'
            frame_style: .primary+1

        /selected:
            content_style: dim bold
            fill_style: '@.panel1+1'
            frame_style: .primary+1

        /active:
            fill_style: 'bold @.primary+3'
            frame_style: '.primary+3'

        /disabled:
            fill_style: '@.panel1-2'
            frame_style: .primary-3
            content_style: .panel1+2

        # Variants

        .big:
            height: 3
    """

    def __init__(
        self,
        content: str,
        on_submit: list[Callable[[Button], bool]] | None = None,
        name: str | None = None,
        **widget_args: Any,
    ) -> None:
        """Initializes the button.

        Args:
            content: The content the button should display.
            on_submit: A list of event callbacks to fire when the button submits.
            name: The name to be used in some parents' serialization result.
        """

        super().__init__(**widget_args)

        self._has_timeout = False

        self.content = content
        self.name = name

        self.on_submit = Event("button submit")
        for callback in on_submit or []:
            self.on_submit += callback

        self.bind("return", wrap_callback(self._visual_submit))

    def _visual_submit(self) -> bool:
        """Animates the active state when the button is submitted using a keyboard."""

        if self._has_timeout:
            return False

        self.state_machine.apply_action("CLICKED")

        self._has_timeout = True
        self.app.timeout(
            150,
            lambda: (
                self.state_machine.apply_action("RELEASED_KEYBOARD"),
                setattr(self, "_has_timeout", False),  # type: ignore
            ),
        )

        return self.on_submit(self)

    def _compute_shrink_width(self) -> int:
        return sum(len(span) for span in self._parse_markup(self.content)) + 4

    def _compute_shrink_height(self) -> int:
        return 1

    def on_click(self, _: MouseAction, __: tuple[int, int]) -> bool:
        """Emits the submit event."""

        self.on_submit(self)

        return True

    def get_content(self) -> list[str]:
        return [self.content]



import string
from typing import Any, Iterator

from slate import Key
from zenith import zml_wrap

from ..enums import MouseAction
from .widget import Widget, to_widget_space

PRINTABLE = [*string.printable]


class Field(Widget):
    """A text field."""

    style_map = Widget.style_map | {
        state: {"cursor": "", "placeholder": ""}
        for state in Widget.state_machine.states
    }

    rules = """
    Field:
        overflow: [auto, auto]
        width: fill
        height: shrink

        frame: [heavy, null, null, null]

        placeholder_style: 'dim italic'
        frame_style: .primary-1

        /hover:
            fill_style: '@.panel1-2'

        /selected|active:
            content_style: ''
            cursor_style: '@.panel1+3'

        /disabled:
            frame_style: .primary-3
    """

    def __init__(
        self,
        value: str = "",
        placeholder: str = "",
        multiline: bool = False,
        scroll_to_cursor: bool = True,
        wrap: bool = False,
        name: str | None = None,
        **widget_args,
    ) -> None:
        self._lines = []

        self.wrap = wrap
        self.placeholder = placeholder
        self.multiline = multiline
        self.scroll_to_cursor = scroll_to_cursor

        self.value = value
        self.name = name

        self.cursor = (0, 0)

        super().__init__(**widget_args)

    @property
    def lines(self) -> list[str]:
        return self._lines

    def _update_lines(self, new: str) -> None:
        if self.wrap:
            self._lines = zml_wrap(new, self._framed_width) or [""]
        else:
            self._lines = new.split("\n")

    @property
    def value(self) -> str:
        return self._value

    @value.setter
    def value(self, new: str) -> None:
        self._update_lines(new)

        self._value = new

    def _get_cursorline(self, x_offset=0, y_offset=0) -> tuple[str, str, str]:
        """Gets the line (left, cursor, right) at the current cursor + given offset."""
        x, y = self.cursor
        x += x_offset
        y += y_offset

        line = self._lines[y]

        left, right = line[:x], line[x + 1 :]
        cursor = line[x] if x < len(line) else ""

        return left, cursor, right

    def _find_word_end(self, line: str, direction: int = 1) -> int:
        """Returns the distance from the next word boundary."""

        # Consistent with unix shell behaviour:
        # * Always delete first char, then remove any non-punctuation
        # Note that the exact behaviour isn't standardized:
        # * Python repl: until change in letter+digit & punctuation
        # * Unix shells: only removes letter+digit
        word_chars = string.ascii_letters + string.digits

        if direction == -1:
            strip_line = line.rstrip(word_chars)
        else:
            strip_line = line.lstrip(word_chars)

        return -direction * (len(strip_line) - len(line)) + direction

    def _compute_shrink_width(self) -> int:
        return max(*[len(line) for line in self.value.split("\n")], 8) + 4

    def _compute_shrink_height(self) -> int:
        return max(len(self.lines), 1)

    def move_cursor(self, x: int = 0, y: int = 0) -> None:
        """Moves the cursor by the given x and y coordinates."""

        self.set_cursor(self.cursor[0] + x, self.cursor[1] + y)

    def set_cursor(self, x: int = 0, y: int = 0) -> None:
        """Sets the cursor to the given coordinates and clamps them."""

        y = max(0, min(len(self.lines) - 1, y))

        line = self.lines[y]

        x = max(0, min(len(line), x))

        self.cursor = x, y

        if not self.scroll_to_cursor:
            return

        # Scroll to cursor
        scroll_offsets = [0, 0]
        frame_start = (self.frame.borders[0] != "", self.frame.borders[1] != "")
        dimensions = (self.computed_width - 1, self.computed_height)

        absolute_cursor = (
            self.cursor[0] + 1 + self.frame.width,
            self.cursor[1] + 1 + self.frame.height,
        )

        for i in range(2):
            opposite_i = 1 * (1 - i)

            box_start = self.scroll[i] + frame_start[i]
            box_end = self.scroll[i] + dimensions[i] - self.has_scrollbar(opposite_i)

            if self.cursor[i] < box_start:
                scroll_offsets[i] = self.cursor[i] - (box_start - frame_start[i])

            elif absolute_cursor[i] > box_end:
                scroll_offsets[i] = absolute_cursor[i] - box_end

        self.scroll = (
            self.scroll[0] + scroll_offsets[0],
            self.scroll[1] + scroll_offsets[1],
        )

    def on_click(self, _: MouseAction, pos: tuple[int, int]) -> bool:
        """Allows the widget to be selected on click."""

        x, y = to_widget_space(pos, self)
        # Account for padding on the left
        self.set_cursor(x - 1, y)

        return True

    def set_line(self, y: int, line: str) -> None:
        """Sets the line at the given y index."""

        self.value = "\n".join(
            [
                *self.lines[:y],
                line,
                *self.lines[y + 1 :],
            ]
        )

    def delete_trailing_newline(self) -> None:
        """Deletes a newline from the end of the current line.

        No-op when y == 0.
        """

        y = self.cursor[1]

        if y == 0:
            return

        parts = self._get_cursorline(y_offset=-1)

        # TODO: Must we copy?
        lines = self.lines.copy()
        lines[y - 1] += lines[y]
        lines.pop(y)

        self.value = "\n".join(lines)

        self.move_cursor(y=-1, x=len("".join(parts)))
        self.scroll = (self.scroll[0], self.scroll[1] - 1)

    def handle_keyboard(self, key: Key) -> bool:
        binds = super().handle_keyboard(key)

        # TODO: Properly handle markup in field value
        if str(key) in "[]":
            return True

        if key == "left":
            self.move_cursor(x=-1)
            return True

        if key == "right":
            self.move_cursor(x=1)
            return True

        if key == "up":
            self.move_cursor(y=-1)
            return True

        if key == "down":
            self.move_cursor(y=1)
            return True

        x, y = self.cursor

        left, cursor, right = self._get_cursorline()

        if key == "alt-left":
            self.move_cursor(x=self._find_word_end(left, direction=-1))
            return True

        if key == "alt-right":
            self.move_cursor(x=self._find_word_end(right + " "))
            return True

        if key == "ctrl-left":
            self.set_cursor(0, y)
            return True

        if key == "ctrl-right":
            self.set_cursor(len(left + cursor + right), y)
            return True

        left += cursor

        if key == "backspace":
            if x == 0:
                self.delete_trailing_newline()
                return True

            self.set_line(y, left[: -1 - len(cursor)] + cursor + right)

            self.move_cursor(x=-1)
            return True

        if key == "ctrl-backspace":
            if x == 0:
                self.delete_trailing_newline()
                return True

            self.set_line(y, cursor + right)
            change = len(left) - 1

            self.move_cursor(x=-change)
            return True

        if key == "alt-backspace":
            if x == 0:
                self.delete_trailing_newline()
                return True

            distance = self._find_word_end(left, direction=-1)

            self.set_line(y, left[:distance] + right)
            self.move_cursor(x=distance)

            return True

        if key == "return":
            if not self.multiline:
                return True

            lines = self.lines

            if cursor != "":
                right = left[-1] + right
                left = left[:-1]

            lines[y] = left

            if len(lines) > y + 1:
                lines.insert(y + 1, right)
            else:
                lines.append(right)

            self.value = "\n".join(lines)

            self.move_cursor(x=-self.cursor[0], y=1)
            return True

        if key in PRINTABLE:
            left = left.removesuffix(cursor)
            self.set_line(y, left + str(key) + cursor + right)

            self.move_cursor(x=1)
            return True

        return binds

    def serialize(self) -> dict[str, Any]:
        if self.name is None:
            raise ValueError(f"field {self!r} cannot be serialized without a name.")

        return {self.name: self.value}

    def get_content(self) -> list[str]:
        value = self.value or self.placeholder

        if not value:
            return [" " + self.styles["cursor"](" ") + " "]

        content_style = self.styles["content"]
        cursor_style = self.styles["cursor"]

        if self.value == "":
            content_style = self.styles["placeholder"]

            if self._selected_index is None:
                cursor_style = content_style

        style = self.styles["placeholder" if not self.value else "content"]
        left, cursor, right = self._get_cursorline()
        if cursor == "":
            cursor = " "

        lines = self.lines
        y = self.cursor[1]

        styled_cursor_line = " " + left + cursor_style(cursor) + "[/]" + right + " "

        return [
            *(f" {line} " for line in lines[:y]),
            styled_cursor_line,
            *(f" {line} " for line in lines[y + 1 :]),
        ]



from .button import Button
from .checkbox import Checkbox
from .container import *
from .dialogue import Dialogue
from .field import Field
from .link import Link
from .text import Text
from .slider import Slider, VerticalSlider
from .progress import Progress
from .dropdown import Dropdown
from .widget import *



from __future__ import annotations

from typing import Any, Iterable, Iterator

from slate import Key, Span

from ..enums import Alignment, Direction, MouseAction, Anchor
from .widget import Widget, _compute, handle_mouse_on_children

__all__ = [
    "Container",
    "Tower",
    "Row",
]


def _align(alignment: Alignment, available: int) -> tuple[int, int]:
    """Returns offset & modulo result for alignment in the available space."""

    available = max(available, 0)

    if available == 1:
        return 0, 0

    if alignment == Alignment.CENTER:
        return divmod(available, 2)

    if alignment == Alignment.END:
        return available, 0

    return 0, 0


class Container(Widget):  # pylint: disable=too-many-public-methods
    """A widget that displays others based on some arrangement algorithm.

    See `Tower` and `Splitter` for the implementations you should use.
    """

    style_map = Widget.style_map | {
        "idle": {"scrollbar_y": ".secondary"},
        "hover": {"fill": Widget.style_map["idle"]["fill"]},
    }

    gap: int | float | None = None
    fallback_gap: int = 0

    def __init__(self, *children: Widget, **widget_args: Any) -> None:
        """Initializes a container.

        Args:
            *children: The children this widget should start with.
        """

        self.children: list[Widget] = []
        self.extend(children)

        super().__init__(**widget_args)

        self._should_layout = True
        self._mouse_target: Widget | None = None
        self._hover_target: Widget | None = None
        self._outer_dimensions = (1, 1)

    @property
    def visible_children(self) -> list[Widget]:
        """Returns all children without the 'hidden' group set."""

        return [widget for widget in self.children if not "hidden" in widget.groups]

    @property
    def direction(self) -> Direction:
        """Returns and sets the current flow direction."""

        return self._direction

    @direction.setter
    def direction(self, new: Direction | str) -> None:
        """Sets the direction setting."""

        if isinstance(new, str):
            new = Direction(new)

        self._direction = new

    @property
    def selectable_count(self) -> int:
        if self.disabled:
            return 0

        return sum(child.selectable_count for child in self.children)

    def __iadd__(self, other: object) -> Container:
        """Adds a widget to the Container.

        Analogous to `.append(other)`.
        """

        if not isinstance(other, Widget):
            raise TypeError(
                f"Can only append widgets to containers, not {type(other)!r}."
            )

        self.append(other)
        return self

    def _is_fill(self, widget: Widget, horizontal: bool) -> bool:
        """Determines whether the widget is fill on the flow axis.

        Args:
            widget: The widget to test.
            horizontal: If set, flow direction is horizontal and we return
                `widget.is_fill_width()`, instead of `is_fill_height`.
        """

        if horizontal:
            return widget.is_fill_width()

        return widget.is_fill_height()

    def _compute_alignment_offsets(
        self,
        child: Widget,
        available_width: int,
        available_height: int,
        horizontal: bool,
    ) -> tuple[int, int, int]:
        """Computes x & y offsets to align by, based on alignment and flow direction.

        Args:
            child: The widget we are aligning. Used to get `computed_width` and
                `computed_height` as needed.
            available_width: The width the widget can be aligned within.
            available_height: The height the widget can be aligned within.
            horizontal: Whether flow direction is horizontal (True) or vertical (False).

        Returns:
            The x offset, y offset and the modulo of the operation that got either x
                offset or y offset, based on `horizontal`.
        """

        if horizontal:
            y, _ = _align(self.alignment[1], available_height - child.computed_height)
            x, extra = _align(self.alignment[0], available_width)

        else:
            x, _ = _align(self.alignment[0], available_width - child.computed_width)
            y, extra = _align(self.alignment[1], available_height)

        return x, y, extra

    # TODO: Revisit shrink calculations
    #
    #       They either add or remove extra space (==frame.dimension), but always the wrong
    #       one of the two.
    def _compute_shrink_width(self) -> int:
        if self.direction is Direction.VERTICAL:
            return self.frame.width + max(
                [child.computed_width for child in self.visible_children], default=0
            )

        gap = 0

        if isinstance(self.gap, int):
            gap = self.gap

        return self.frame.width + sum(
            child.computed_width + gap for child in self.visible_children
        ) - gap

    def _compute_shrink_height(self) -> int:
        if self.direction is Direction.HORIZONTAL:
            return self.frame.height + max(
                [child.computed_height for child in self.visible_children], default=0
            )

        gap = 0

        if isinstance(self.gap, int):
            gap = self.gap

        return self.frame.height + sum(
            child.computed_height + gap for child in self.visible_children
        ) - gap

    @property
    def selected(self) -> Widget | None:
        return self._selected

    def select(self, index: int | None = None) -> int | None:
        if index == 0 or self.selectable_count == 0:
            return index

        possible_index = index

        if index is None:
            if self._selected is not None:
                self._selected.select(None)

            self._selected_index = None
            return None

        if self._selected is not None:
            self._selected.select(None)

        for widget in self.children:
            val = widget.select(index)
            index = val

            if index <= 0:
                self._selected = widget
                self._selected_index = possible_index
                self.state_machine.apply_action("SELECTED")
                break

        else:
            # We did not contain index, so we can unselect
            self._selected_index = None
            self.state_machine.apply_action("UNSELECTED")

        return max(index, 0)

    def serialize(self) -> dict[str, str]:
        data: dict[str, str] = {}

        for child in self.children:
            data.update(**child.serialize())

        return data

    def insert(self, index: int, widget: Widget) -> None:
        """Inserts a widget.

        Analogous to `list.insert`.
        """

        self.children.insert(index, widget)
        widget.parent = self
        self._should_layout = True

    def append(self, widget: Widget) -> None:
        """Adds a new widget setting its parent attribute to self.

        Analogous to `list.append`.
        """

        self.insert(len(self.children), widget)

    def extend(self, widgets: Iterable[Widget]) -> None:
        """Extends our children by the given iterable.

        Analogous to `list.extend`.
        """

        for widget in widgets:
            self.append(widget)

    def remove(self, widget: Widget) -> None:
        """Removes a widget from self, resetting its parent attribute to None.

        Analogous to `list.remove`.
        """

        self.children.remove(widget)

        widget.parent = None
        self._should_layout = True

        if self._mouse_target is widget:
            self._mouse_target = None

    def pop(self, index: int = -1) -> Widget:
        """Pops a widget from our children.

        Analogous to `list.pop`.

        Returns:
            The widget that was just removed.
        """

        widget = self.children[index]
        self.remove(widget)

        return widget

    def clear(self) -> None:
        """Removes all widgets.

        Analogous to `list.clear`.
        """

        for widget in self.children.copy():
            self.remove(widget)

    def update_children(self, widgets: Iterable[Widget]) -> None:
        """Updates our children to the given iterable.

        Analogous to `list.update`, doesn't use `update` because it is already a widget
        method.
        """

        self.clear()
        self.extend(widgets)

    def replace(self, current: Widget, new: Widget, *, offset: int = 0) -> None:
        """Replaces `current` in children to `new`."""

        index = self.children.index(current)

        self.children[index + offset] = new
        new.parent = self

    def move_by(self, x: int, y: int) -> None:
        """Moves the widget (and all its children) to the given position."""

        super().move_by(x, y)

        for child in self.children:
            child.move_by(x, y)

    def arrange(  # pylint: disable=too-many-locals,too-many-statements,too-many-branches
        self, x: int, y: int
    ) -> None:
        """Arranges the widget's children according to its flow.

        Args:
            x: The origin's horizontal coordinate.
            y: The origin's vertical coordinate.
        """

        children = self.visible_children

        layouted_children = [
            child for child in self.visible_children if child.anchor == Anchor.NONE
        ]
        layouted_count = len(layouted_children)

        width = self._framed_width - self.has_scrollbar(1)
        height = self._framed_height - self.has_scrollbar(0)

        is_horizontal = self.direction == Direction.HORIZONTAL

        fill_buffer = width if is_horizontal else height
        fill_count = 0

        for child in layouted_children:
            if self._is_fill(child, is_horizontal):
                fill_count += 1
                continue

            child.compute_dimensions(width, height)

            if is_horizontal:
                fill_buffer -= child.computed_width
            else:
                fill_buffer -= child.computed_height

        gap = self.gap
        gap_extra = 0

        if gap is None:
            gap = 0

            if fill_count == 0:
                gap, gap_extra = divmod(fill_buffer, max(layouted_count, 1))
                
                if gap < 0:
                    gap = 0
                    gap_extra = 0

        else:
            fill_buffer -= (layouted_count - 1) * gap

        fill_size, fill_remainder = divmod(fill_buffer, max(fill_count, 1))

        origin = x, y

        t_width, t_height = self.terminal.size
        t_ox, t_oy = self.terminal.origin

        s_start, s_end = [list(val) for val in self.inner_rect]

        for child in children:
            if self._is_fill(child, is_horizontal):
                fill_extra = 1 if fill_remainder > 0 else 0

                if is_horizontal:
                    child.compute_dimensions(fill_size + fill_extra, height)
                else:
                    child.compute_dimensions(width, fill_size)

                fill_remainder -= 1

            if child.anchor == Anchor.SCREEN:
                child.compute_dimensions(width, height)

                offset = list(child.offset)

                if offset[0] == "end":
                    offset[0] = self.computed_width - child.computed_width

                elif child.offset[0] < 0:
                    offset[0] -= -offset[0]

                if offset[1] == "end":
                    offset[1] = self.computed_height - child.computed_height

                elif child.offset[1] < 0:
                    offset[1] -= -offset[1]

                child.move_to(x + offset[0], y + offset[1])

            elif child.anchor == Anchor.PARENT:
                child.compute_dimensions(width, height)

                offset = list(child.offset)

                if offset[0] == "end":
                    offset[0] = t_width - child.computed_width

                elif offset[0] < 0:
                    offset[0] -= -offset[0]

                if offset[1] == "end":
                    offset[1] = t_height - child.computed_height

                elif offset[1] < 0:
                    offset[1] -= -offset[1]

                child.move_to(t_ox + offset[0], t_oy + offset[1])

            else:  # child.anchor == Anchor.NONE
                # TODO: Some of these could be computed out of loop.
                if is_horizontal:
                    align_x, align_x_extra = _align(
                        self.alignment[0],
                        gap,
                    )
                    align_y, align_y_extra = _align(
                        self.alignment[1],
                        height - child.computed_height,
                    )

                    child.move_to(
                        x + align_x + align_x_extra, y + align_y + align_y_extra
                    )
                    x += child.computed_width + gap + (1 * gap_extra > 0)

                else:
                    align_x, align_x_extra = _align(
                        self.alignment[0],
                        width - child.computed_width,
                    )
                    align_y, align_y_extra = _align(
                        self.alignment[1],
                        gap,
                    )

                    child.move_to(
                        x + align_x + align_x_extra, y + align_y + align_y_extra
                    )
                    y += child.computed_height + gap + (1 * gap_extra > 0)

                gap_extra -= 1

                clip_start, clip_end = [0, 0], [0, 0]

                c_start, c_end = child.outer_rect

                if c_start[0] < s_start[0]:
                    clip_start[0] = s_start[0] - c_start[0]

                if c_start[1] < s_start[1]:
                    clip_start[1] = s_start[1] - c_start[1]

                if s_end[0] < c_end[0]:
                    clip_end[0] = c_end[0] - s_end[0]

                if s_end[1] < c_end[1]:
                    clip_end[1] = c_end[1] - s_end[1]

                child.clip(clip_start, clip_end)

        self._outer_dimensions = [
            self._compute_shrink_width(),
            self._compute_shrink_height(),
        ]

    def get_content(self) -> list[str]:
        """Calls our `arrange` method and returns a single empty line."""

        # layout_state = self._as_layout_state()

        # if layout_state != self._layout_state:
        start_x = self.position[0] + (self.frame.left != "")
        start_y = self.position[1] + (self.frame.top != "")

        if self._should_layout:
            self.arrange(start_x - self.scroll[0], start_y - self.scroll[1])
        # self._layout_state = layout_state

        return [""]

    def handle_keyboard(self, key: Key) -> bool:
        if self.selected is not None and self.selected.handle_keyboard(key):
            return True

        idx = self._selected_index or 0

        if key in ("left", "up", "shift-tab"):
            self.select(idx - 1)

        if key in ("right", "down", "tab"):
            self.select(idx + 1)

        return super().handle_keyboard(key)

    def handle_mouse(self, action: MouseAction, position: tuple[int, int]) -> bool:
        result, mouse_target, hover_target = handle_mouse_on_children(
            action,
            position,
            self._mouse_target,
            self._hover_target,
            self.visible_children,
        )

        self._mouse_target = mouse_target
        self._hover_target = hover_target

        if result:
            index = 0

            for child in self.children:
                if child is mouse_target:
                    index += child.selected_index or 1
                    self.select(index)

                    return True

                index += child.selectable_count

            return True

        # There was a click, but it didn't get handled; likely clicked on a
        # non-selectable, or empty space.
        if "click" in action.value:
            self.select(None)

        return super().handle_mouse(action, position)

    def drawables(self) -> Iterator[Widget]:
        yield from super().drawables()

        for widget in sorted(self.children, key=lambda w: w.layer):
            yield from widget.drawables()

    def build(
        self, *, virt_width: int | None = None, virt_height: int | None = None
    ) -> list[tuple[Span, ...]]:
        my_virt_width, my_virt_height = self._outer_dimensions

        return super().build(virt_width=my_virt_width, virt_height=my_virt_height)


class Tower(Container):
    """Arranges widgets in a vertical tower.

    ```
    +----------------+ <- alignment=("center", "center")
    |                |
    |     +---+      |
    |     |   | <------- width=0.5
    |     +---+      |
    |   +--------+   |
    |   |        | <---- width=1.0
    |   +--------+   |
    |    +-----+     |
    |    |     | <------ width=7
    |    +-----+     |
    |                |
    +----------------+
    ```
    """

    _direction = Direction.VERTICAL


class Row(Container):
    """Arranges widgets in a horizontal row.

    ```
    +----------------------------+ <- alignment=("start", "center")
    |                            |
    |      +--------++----------+|
    |+----+|        ||          ||
    ||    ||        ||          ||
    |+----+|        ||          ||
    |  ^   +--------++----------+|
    |  |          ^         ^    |
    +--|----------|---------|----+
       |_ width=6 |         |
                  |_ width=0.3
                            |_ width=0.3
    ```
    """

    _direction = Direction.HORIZONTAL



from __future__ import annotations

from typing import Any

from .widget import Widget


class Progress(Widget):
    """A simple, non-interactable progress bar."""

    rail: str = "─"
    filled: str = "━"
    show_percentage: bool = True

    style_map = Widget.style_map | {
        state: {"percentage": ""} for state in Widget.state_machine.states
    }

    rules = """
    Progress:
        height: 1

        frame_style: '.panel1'
        content_style: '.primary'

        .minimal:
            show_percentage: false

        .tall:
            rail: '█'
            filled: '█'

            percentage_style: '@.primary .panel1'
            frame_style: '.panel1-1'
    """

    def __init__(self, value: float = 0.0, **widget_args: Any) -> None:
        super().__init__(**widget_args)

        self.value = value

    @property
    def value(self) -> float:
        """Returns the internal value."""

        return self._value

    @value.setter
    def value(self, new: float) -> None:
        """Sets the internal value."""

        self._value = new

    def get_content(self) -> list[str]:
        filled = round(self.computed_width * self.value)

        style = self.styles["content"]

        if not self.show_percentage:
            return [
                self.styles["content"](style(filled * self.filled))
                + self.styles["frame"]((self.computed_width - filled) * self.rail)
            ]

        percentage = f" {self.value * 100:.0f}% "
        filled -= len(percentage)

        return [
            self.styles["content"](filled * self.filled)
            + self.styles["percentage"](percentage)
            + self.styles["frame"]((self.computed_width - filled) * self.rail)
        ]



from __future__ import annotations

from io import StringIO
from contextlib import contextmanager

from celadon import Application, Page, Tower, Row, Text, Widget

from slate import Span, Terminal, Color
from zenith import zml_get_spans


class SizedTerminal(Terminal):
    stream = StringIO()

    @property
    def size(self) -> tuple[int, int]:
        return 80, 24

    @contextmanager
    def no_echo(self):
        try:
            yield
        finally:
            pass


def _format_lines(lines: list[tuple[Span]]) -> str:
    return "\n" + ",\n".join(
        "(" + ", ".join(map(repr, map(str, line))) + ")" for line in lines
    )


def _as_spans(text: list[tuple[str]]) -> list[tuple[Span]]:
    def _to_span(item: str | Span) -> Span:
        if isinstance(item, Span):
            return item

        return Span(item)

    output = []
    length = len(text)

    for i, line in enumerate(text):
        output.append(tuple(map(_to_span, filter(lambda item: item != "", line))))

    return output


def apply_rules(widget: Widget, rules: str | None = None) -> Widget:
    with Application("Test Runner", terminal=SizedTerminal()) as app:
        app.rule(
            "*",
            content_style="",
            frame_style="",
            fill_style="",
            scrollbar_x_style="",
            scrollbar_y_style="",
        )

        app += Page(Tower(widget), rules=(rules or ""))
        app.timeout(0, lambda: app.stop())

    return widget


def test_widget_alignment() -> None:
    w = apply_rules(
        Text("hello"),
        """
    Text:
        width: 20
        height: 5

        frame: ascii_x
    """,
    )

    w.alignment = ["start", "start"]

    assert (output := w.build()) == _as_spans(
        [
            ("X", "------------------", "X"),
            ("|", "hello             ", "|"),
            ("|", "                  ", "|"),
            ("|", "                  ", "|"),
            ("X", "------------------", "X"),
        ]
    ), _format_lines(output)

    w.alignment = ["center", "center"]

    assert (output := w.build()) == _as_spans(
        [
            ("X", "------------------", "X"),
            ("|", "                  ", "|"),
            ("|", "       ", "hello", "      ", "|"),
            ("|", "                  ", "|"),
            ("X", "------------------", "X"),
        ]
    ), _format_lines(output)

    w.alignment = ["start", "end"]

    assert (output := w.build()) == _as_spans(
        [
            ("X", "------------------", "X"),
            ("|", "                  ", "|"),
            ("|", "                  ", "|"),
            ("|", "hello             ", "|"),
            ("X", "------------------", "X"),
        ]
    ), _format_lines(output)


def test_widget_scrolling() -> None:
    w = apply_rules(
        Text(
            "\n".join("".join(map(lambda i: str(i % 10), range(20))) for _ in range(20))
        ),
        """
        Text:
            width: 20
            height: 10
            overflow: [scroll, scroll]

            frame: ascii_x
        """,
    )

    assert (output := w.build()) == _as_spans(
        [
            ("X", "------------------", "X"),
            ("|", "01234567890123456", "#", "|"),
            ("|", "01234567890123456", "#", "|"),
            ("|", "01234567890123456", "#", "|"),
            ("|", "01234567890123456", "|", "|"),
            ("|", "01234567890123456", "|", "|"),
            ("|", "01234567890123456", "|", "|"),
            ("|", "01234567890123456", "|", "|"),
            ("|", "################- ", "|"),
            ("X", "------------------", "X"),
        ]
    ), _format_lines(output)

    w.scroll_to(y=-1)

    assert (output := w.build()) == _as_spans(
        [
            ("X", "------------------", "X"),
            ("|", "01234567890123456", "|", "|"),
            ("|", "01234567890123456", "|", "|"),
            ("|", "01234567890123456", "|", "|"),
            ("|", "01234567890123456", "|", "|"),
            ("|", "01234567890123456", "|", "|"),
            ("|", "01234567890123456", "#", "|"),
            ("|", "01234567890123456", "#", "|"),
            ("|", "################- ", "|"),
            ("X", "------------------", "X"),
        ]
    ), _format_lines(output)

    w.scroll_to(x=5, y=7)

    assert (output := w.build()) == _as_spans(
        [
            ("X", "------------------", "X"),
            ("|", "34567890123456789", "|", "|"),
            ("|", "34567890123456789", "|", "|"),
            ("|", "34567890123456789", "#", "|"),
            ("|", "34567890123456789", "#", "|"),
            ("|", "34567890123456789", "#", "|"),
            ("|", "34567890123456789", "|", "|"),
            ("|", "34567890123456789", "|", "|"),
            ("|", "-################ ", "|"),
            ("X", "------------------", "X"),
        ]
    ), _format_lines(output)


def test_widget_slice_line() -> None:
    w = Widget()

    spans = zml_get_spans("[blue]tes[@red]t[/ blue] content")
    sliced = w._slice_line(spans, 0, 3)
    assert sliced == (Span("tes", reset_after=True, foreground=Color(rgb=(0, 0, 255))),)


def test_widget_selection() -> None:
    w = Widget()
    assert w._selected_index is None
    assert w.state == "idle"

    w.select(1)
    assert w._selected_index == 0
    assert w.selected is w
    assert w.state == "selected"

    target = Widget()

    w = Tower(Widget(), Widget(), target, Widget())

    w.select(0)
    assert w._selected_index is None

    w.select(3)
    assert w._selected_index == 3
    assert w.selected is target

    nested_target = Widget()
    target = Tower(Widget(), nested_target, Widget())
    outer_target = Widget()
    w = Tower(Widget(), outer_target, target, Widget())

    w.select(4)
    assert w.selected is target
    assert w.selected.selected is nested_target
    assert w.state == "selected"

    w.select(2)
    assert w.selected is outer_target
    assert w.state == "selected"



from __future__ import annotations

from typing import Any

from slate import Event

from ..enums import MouseAction

from .text import Text
from .button import Button
from .container import Tower
from .widget import wrap_callback, _compute


class Dropdown(Tower):
    """A dropdown inspired by HTML's select, taking Button objects."""

    indicators = ["▼", "▲"]

    rules = """
    Dropdown:
        width: shrink

        \> Button:
            width: null
    """

    # Display over other content
    layer = 1

    # Stack without gaps
    gap = 0
    fallback_gap = 0

    def __init__(
        self,
        title: str,
        items: list[Button] | None = None,
        name: str | None = None,
        **widget_args: Any,
    ) -> None:
        super().__init__(*(items or []), **widget_args)

        self.value = title
        self.label = title
        self.name = name
        self.is_open = False

        self._trigger = Button("")
        self.insert(0, self._trigger)
        self.insert(1, Button(title, on_submit=[self._select_option]))

        self._trigger.on_submit += wrap_callback(self.toggle)
        self.height = 1

        def _update_trigger_content() -> None:
            self._trigger.content = f"{self.label} {self.indicators[self.is_open]}"

        def _close_on_deselect() -> None:
            if self.selected_index is None:
                self.close()

        self.pre_content += wrap_callback(_update_trigger_content)
        self.pre_build += wrap_callback(_close_on_deselect)

    def serialize(self) -> dict[str, Any]:
        if self.name is None:
            raise ValueError(f"dropdown {self!r} cannot be serialized without a name.")

        return {self.name: self.value}

    def _select_option(self, button: Button) -> bool:
        """Selects one of the option buttons and updates our state."""

        if button not in self.children:
            raise ValueError("can't select {button!r} since it's not part of {self!r}")

        self.close()
        self.value = button.name or button.content
        self.label = button.content

        return True

    def open(self) -> None:
        """Opens the dropdown."""

        self.is_open = True

    def close(self) -> None:
        """Closes the dropdown."""

        self.is_open = False

    def toggle(self) -> bool:
        """Toggles the dropdown."""

        if self.is_open:
            self.close()
        else:
            self.open()

        return self.is_open

    def append(self, widget: Widget) -> None:
        """Let's any new button child select options."""

        if not isinstance(widget, (Button, Text)):
            raise TypeError(
                "dropdown can only take Buttons and Texts, got {type(widget)!r}."
            )

        super().append(widget)

        if isinstance(widget, Button):
            widget.on_submit += self._select_option

    def drawables(self) -> Iterator[Widget]:
        """Yields children based on whether the dropdown is open."""

        if not self.is_open:
            yield from (self, self._trigger)
            return

        yield from super().drawables()

    def _compute_shrink_width(self) -> int:
        return max(len(widget.content) for widget in self.children) + 4

    @property
    def visible_children(self) -> list[Widget]:
        """Yields visible children based on whether the dropdown is open."""

        if self.is_open:
            return [widget for widget in self.children if not "hidden" in widget.groups]

        return super().visible_children

    def build(
        self, *, virt_width: int | None = None, virt_height: int | None = None
    ) -> list[tuple[Span, ...]]:
        """Fakes `computed_height` while open to allow for dropdown behavior."""

        self.computed_height = len(self.children) if self.is_open else 1

        return super().build(virt_width=virt_width, virt_height=virt_height)
